#!/usr/bin/perl
#
# e2ctrl.pl Version 0.12 alpha: Enigma 2 remote control interface
#
# Author/Copyright (c) 2010 - 2011: Christian Birchinger <cbirchinger@netswarm.net>
# All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
# FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
use utf8;
use Encode;
use strict;
use Glib ':constants';
use Gtk2 '-init';
use Gtk2::TrayIcon;
use Gtk2::SimpleList;
use Gtk2::Gdk::Keysyms;
use LWP::UserAgent;
use XML::Simple;
use Env qw(HOME);
use Fcntl qw(LOCK_EX LOCK_NB);
use Cwd;
use Getopt::Std;
#use Data::Dumper;

#---------------------------------------------------------------------------------------

my $appname = "e2ctrl";
my $appver = "0.12alpha";
my $iconname = "${appname}.png";
my $remoteame = "e2ctrl-rmtable.txt";
my $zapmethod = 'direct'; # 'remote' (stays in sync with the TV interface but sucks) or 'direct' (faster)
my $startupdir = getcwd();
my @pixmappaths = ($startupdir, "$ENV{HOME}/.${appname}", "/usr/share/pixmaps", "/usr/local/share/pixmaps");
my @datapaths = ($startupdir, "$ENV{HOME}/.${appname}", "/usr/share/${appname}", "/usr/local/share/${appname}");
my $http_timeout = 10;
my $debug = 0;
my $nodaemon = 0;
my $useproxy = 1;

my $pwrstate = "offline";
my %e2about;
my $bouquets;
my $services;
my @services_index;
my $current_channel = 1;
my $current_channel_info;
my $last_tooltip_channel;

my %power_codes = (
	'standby' => '0',
	'deep-standby' => '1',
	'reboot' => '2',
	'restart-enigma' => '3'
);

#-[ Checking external data ]------------------------------------------------------------

my $iconfile = get_location($iconname, \@pixmappaths);
abort("Unable to open icon ${iconname}.") if (!$iconfile);

my $remotefile = get_location($remoteame, \@datapaths);
abort("Unable to open remote control code file ${remotefile}.") if (!$remotefile);
my %remote_codes = get_remote_codes();

#-[ Argument parsing ]------------------------------------------------------------------
our($opt_r, $opt_h, $opt_f, $opt_t, $opt_c, $opt_d, $opt_v, $opt_n);
getopts('hft:d:vn');

if ($opt_v) {
	print "${appname} ${appver}\n";
	exit 0;
}

display_help() if ($opt_h || ! $ARGV[0]);

my $host = $ARGV[0];
my $baseurl = "http://${host}/web/";

$nodaemon = 1 if ($opt_f);
$http_timeout = $opt_c if ($opt_c =~ /^\d{1,3}$/ && $opt_c > 4);

$debug = $opt_d if ($opt_d =~ /^\d$/);
debug("Debug level is set to: $debug", 1);

$useproxy = 0 if ($opt_n);
#---------------------------------------------------------------------------------------

debug("http request timeout is: $http_timeout", 2);

my $ua = LWP::UserAgent->new( env_proxy => $useproxy,
                              keep_alive => 1,
                              timeout => $http_timeout, );

daemonize() if ($nodaemon == 0);
$0 = $appname if ($^O eq 'linux'); # Setting $0 fails on FreeBSD and Windows

#-[ GTK 2 User interface ]--------------------------------------------------------------

Gtk2->init;
my $icon = Gtk2::Image->new_from_file($iconfile);
my $eventbox = Gtk2::EventBox->new;
$eventbox->add( $icon );

my $tray = Gtk2::TrayIcon->new( $appname );
$tray->add( $eventbox );

my $tooltip = Gtk2::Tooltips->new;
$tooltip->set_tip( $tray, "${appname} ${appver} (initialising connection to ${host})" );

$eventbox->signal_connect( 'button_press_event', \&mouse_click );
$eventbox->signal_connect( 'scroll_event', \&mouse_wheel );

$tray->show_all;

# Channel Window ======================================================================================
my $channel_input;

my $channel_window = Gtk2::Window->new( 'toplevel' );
$channel_window->set_position( 'center' );
$channel_window->set_title ( 'Channels' );
$channel_window->set_default_size (500, 600);
$channel_window->set_default_icon_from_file($iconfile);

my $channel_window_treeview = Gtk2::SimpleList->new (
	'Nr' => 'int',
	'Name' => 'text',
	'EPG Now' => 'text',
	'Runtime' => 'text'
);

my $channel_window_vbox = Gtk2::VBox->new( 0, 6 );
$channel_window->add($channel_window_vbox);
$channel_window_vbox->set_border_width(2);

# Channel Entry Box  ------------------------------------------------------------
my $channel_window_entry_hbox = Gtk2::HBox->new( 0, 6 );
$channel_window_vbox->pack_start($channel_window_entry_hbox,0,0,2);

my $label_entry = Gtk2::Label->new('Enter channel number:');
$channel_window_entry_hbox->pack_start($label_entry, 0, 0, 4 );

my $channel_window_entry = Gtk2::Entry->new();
$channel_window_entry->signal_connect (changed => sub{ $channel_input = $channel_window_entry->get_text; } );
$channel_window_entry_hbox->pack_start($channel_window_entry, 1, 1, 0 );
$channel_window_entry->signal_connect(activate => sub{
	zap_channel_number($channel_input);
	$channel_window_entry->set_text("");
	$channel_window_treeview->select($current_channel - 1);
});

my $channel_window_entry_button = Gtk2::Button->new_with_label('Send');
$channel_window_entry_hbox->pack_start( $channel_window_entry_button, 0, 0, 2 );
$channel_window_entry_button->signal_connect( clicked => sub{
	zap_channel_number($channel_input);
	$channel_window_entry->set_text("");
	$channel_window_treeview->select($current_channel - 1);
});
# -------------------------------------------------------------------------------

# Scrollable Simple List --------------------------------------------------------
my $channel_window_treeview_scroll = Gtk2::ScrolledWindow->new;;
$channel_window_treeview_scroll->set_policy ('never', 'always'); # horizontal, vertical (always, never, automatic)
$channel_window_vbox->pack_start($channel_window_treeview_scroll, 1, 1, 4 );

$channel_window_treeview->set_headers_visible(TRUE);
$channel_window_treeview_scroll->add($channel_window_treeview);

$channel_window_treeview->get_selection()->signal_connect(changed => sub {
	my ($selection) = @_;
	my ($model, $iter) = $selection->get_selected;
	return if (!$iter);
	$selection->get_tree_view->scroll_to_cell ($model->get_path ($iter)); # Scroll to selected entry
	my ($channum, $channame) = $model->get($iter);
	debug("New channel selected ${channame} (${channum}).", 2);
	zap_channel_number($channum);
});

# Close Button ------------------------------------------------------------------
my $channel_window_close_hbox= Gtk2::HBox->new( 0, 6 );

my $close_button = Gtk2::Button->new_from_stock('gtk-close');
$channel_window_close_hbox->pack_end( $close_button, 0, 0, 6 );
$close_button->signal_connect( clicked => sub{ $channel_window->hide; } );

$channel_window_vbox->pack_start($channel_window_close_hbox,0,0,2);
# -------------------------------------------------------------------------------

$channel_window->signal_connect(delete_event => sub {	$channel_window->hide; return 1; });

$channel_window->signal_connect (key_press_event => sub {
		my ($widget, $event) = @_;
		return  unless $event->keyval == $Gtk2::Gdk::Keysyms{KEY_Escape};
		$channel_window->hide;
		return 1;
	});

# Set focus on the string input widget by default
$channel_window->signal_connect ('focus-in-event' => sub{ $channel_window_entry->grab_focus() });

# ==========================================================================================================

toggle_sensitive_widgets(0);

Glib::Timeout->add (10000, \&background_10s, undef, G_PRIORITY_DEFAULT_IDLE);
Glib::Timeout->add (59000, \&background_update_services, undef, G_PRIORITY_DEFAULT_IDLE);
Glib::Timeout->add (71000, \&background_update_current_channel, undef, G_PRIORITY_DEFAULT_IDLE);
background_10s(); # Also run once at the beginning

Gtk2->main;
debug("Exiting normaly.", 1);
exit 0;

#---------------------------------------------------------------------------------------

sub toggle_sensitive_widgets {
	my ($value) = @_;

	my @widgetlist = ($channel_window_treeview_scroll, $channel_window_entry_hbox);

	for my $widget (@widgetlist) {
		if ($value != 1 && $value != 0) {
			if ($widget->sensitive()) {
				$value = 0;
			} else {
				$value = 1;
			}
		}
		debug ( "Setting widgets sensitivity to: $value", 2 );
		$widget->set_sensitive($value);
	}
}

sub update_current_channel {
	$current_channel_info = get_xml_from_e2cmd( 'getcurrent' );
	if ( defined($current_channel_info) ) {
		my $i=1;
		for my $s (@services_index) {
			if ( $s->{'e2eventservicereference'} eq $current_channel_info->{'e2service'}->{'e2servicereference'} ) {
				debug ( "Current active channel number is: ${i}", 3 );
				$current_channel = $i;
			}
			$i++;
		}
	}
}

sub update_services {
	debug ( "Updating service database.", 2 );

	$services = get_services();
	undef @services_index;

	if ( !defined($services) ) {
		return FALSE;
	}

	for my $entry ( @{$services} ) {
		my $subservices = get_xml_from_e2cmd( 'epgnow?bRef=' . $entry->{'e2servicereference'} );

		if ( defined($subservices) ) {
			$entry->{'e2servicetree'} = $subservices->{'e2event'};

			for my $service ( @{$entry->{'e2servicetree'}} ) {
				push @services_index, $service;
			}
		} else {
			undef $entry->{'e2servicetree'};
		}
	}

	update_channel_window_treeview();

	return TRUE;
}

sub get_services {
	my ($reference) = @_;
	my $xmlresult;

	if ($reference) {
		$xmlresult = get_xml_from_e2cmd( 'getservices?sRef=' . $reference );
	} else {
		$xmlresult = get_xml_from_e2cmd( 'getservices' );
	}

	# Workaround: result needs to be an array reference which is not the
	# case if theres only one entry in the XML.
	if (ref($xmlresult->{'e2service'}) eq 'HASH') {
		my @array;
		$array[0] = $xmlresult->{'e2service'};
		return \@array;
	}

	return $xmlresult->{'e2service'};
}

sub background_update_services {
	update_services() if ( $pwrstate eq 'running' );
	return 1;
}

sub background_update_current_channel {
	update_current_channel() if ( $pwrstate eq 'running' );
	return 1;
}

sub background_10s {
	my $update_tooltip = 0;
	my $channel_tooltip = '';

	my $oldpwrstate = $pwrstate;
	$pwrstate = get_power_state();
	$update_tooltip = 1 if ($pwrstate ne $oldpwrstate);

	if ( !%e2about && ($pwrstate eq 'running' || $pwrstate eq 'standby') ) {
		debug ( "Requesting about information from device.", 1 );
		my $xmlabout = get_xml_from_e2cmd( "about" );
		if (defined($xmlabout)) {
			%e2about = %{$xmlabout->{'e2about'}};
			debug ( "Device model is $e2about{'e2model'}.", 1 );
		}
	}

	# Box recently turned on
	if ( $pwrstate eq 'running' && $oldpwrstate ne 'running' ) {
		update_services();
		update_current_channel();
		toggle_sensitive_widgets(1);
	}

	# Box recently turned off or put into standby
	if ( $pwrstate ne 'running' && $oldpwrstate eq 'running' ) {
		toggle_sensitive_widgets(0);
	}

	if ( $pwrstate eq 'running' ) {
		if ($last_tooltip_channel ne $current_channel || $update_tooltip) {
			debug ( "Channel changed, queuing tooltip update.", 2 );
			$channel_tooltip = " / " . $services_index[$current_channel - 1]->{'e2eventservicename'};
			$update_tooltip = 1;
		}
		update_channel_selection();
	}

	if ( $update_tooltip ) {
		debug ( "Updating tooltip info.", 2 );
		$tooltip->set_tip( $tray, "$e2about{'e2model'}(${pwrstate})${channel_tooltip}" );
		$last_tooltip_channel = $current_channel;
	}

	return 1;
}

sub update_channel_selection {
	my ($selected) = $channel_window_treeview->get_selected_indices;
	if ( !defined($selected) || $current_channel != ($selected+1) ) {
 		debug ( "Active (" . $current_channel . ") and selected channel (" . ($selected+1) . ") differ. Updating listview selection.", 1);
		$channel_window_treeview->select($current_channel - 1);
	}
}

sub get_power_state {
	my $xmlresult = get_xml_from_e2cmd( 'powerstate' );
	my $state;

	if (defined($xmlresult)) {
		my $result = $xmlresult->{'e2instandby'};
		$result =~ s/^\s+//; $result =~ s/\s+$//; # Fix broken XML output from 'powerstate'
		if ($result eq 'false' ) {
			$state = 'running';
		} elsif ( $result eq 'true' ) {
			$state = 'standby';
		} else {
			$state = 'unknown';
		}
	} else {
		$state = 'offline';
	}
	return $state;
}

sub get_xml_from_e2cmd {
	my ($url) = @_;

	debug( "Requesting URL ${baseurl}${url} from device.", 2 );

	my $response = $ua->get( $baseurl . $url );

	my $webcontent = $response->{'_content'};
	my $rc = $response->{'_rc'};
	if ( $rc != 200 ) {
		debug ( "WARNING: Received a HTTP error: $rc.", 1 );
		return undef;
	}
	Encode::_utf8_on($webcontent);
	return XMLin($webcontent);
}

sub powerstate_command {
	my ($command) = @_;
	my $code = $power_codes{ $command };
	debug ( "Power state command: ${command}(${code})", 1 );
	get_xml_from_e2cmd( "powerstate?newstate=" . $code );
}

sub remote_command {
	my ($command) = @_;
	my $code = $remote_codes{ $command };
	if (!defined($code)) {
		debug("WARNING: no code for remote control command \"$command\" found.", 1);
		return
	}
	debug ( "Executing remote control command: ${command}(${code})", 1 );
	get_xml_from_e2cmd( "remotecontrol?command=" . $code );
}

sub mouse_click {
	debug ( "Mouse button (" . $_[ 1 ]->button . ") pressed.", 2 );

	if ( $_[ 1 ]->button == 1 ) {
		toggle_channel_win();
	}	elsif ( $_[ 1 ]->button == 2 ) {
		remote_command('power');
	}	elsif ( $_[ 1 ]->button == 3 ) {
		&popup_menu;
	}	elsif ( $_[ 1 ]->button == 8 ) {
		previous_channel();
	}	elsif ( $_[ 1 ]->button == 9 ) {
		next_channel();
	} else {
		#
	}
	return 1;
}

sub toggle_channel_win {
	if ($channel_window->is_active) {
		$channel_window->hide;
	} else {
		$channel_window->show_all;
		$channel_window->present;
	}
}

sub mouse_wheel {
	debug ( "Mouse wheel scrolled " . $_[ 1 ]->direction . ".", 2 );

	if ( $_[ 1 ]->direction eq 'up' ) {
		next_channel();
	} elsif ( $_[ 1 ]->direction eq 'down' ) {
		previous_channel();
	} else {
		#
	}
	return 1;
}

sub get_remote_codes {
	my $line;
	my %cmdtable;
	my $linepos = 0;

	open (RMCODES, "$remotefile");
	while ($line = <RMCODES>) {
		++$linepos;
		next if $line =~ /^(#|\s*$)/;
		(my $nr, my $cmd) = $line =~ /\s*(\d+)\s+[Kk][Ee][Yy]\s+"([^"]+)"/;

		if (defined $cmd && $nr =~ /\d+/) {
			debug( "Add remote control entry: " . lc($cmd) . " = " . $nr, 4);
			$cmdtable{ lc($cmd) } = $nr;
		}	else {
			print "WARNING: parse error while reading $remotefile in line $linepos:\n";
			print "=> $line";
		}
	}
	close RMCODES;

	return %cmdtable;
}

sub popup_menu {
	debug ( "Opening right mouse button popup menu.", 1 );

	my $menu = Gtk2::Menu->new();

	my $menu_channelwin = Gtk2::ImageMenuItem->new_with_label( "Channel Window" );
	$menu_channelwin->set_image( Gtk2::Image->new_from_stock( 'gtk-index', 'menu' ) );
	$menu_channelwin->signal_connect( activate => sub { $channel_window->show_all } );
	$menu->append( $menu_channelwin );

	my $menu_audiowin = Gtk2::ImageMenuItem->new_with_label( "Select Audio Track" );
	$menu_audiowin->set_image( Gtk2::Image->new_from_icon_name( 'sound', 'menu' ) );
	$menu_audiowin->signal_connect( activate => \&open_audio_window );
	$menu->append( $menu_audiowin );
	$menu_audiowin->set_sensitive(0) if ($pwrstate ne 'running');

	# Power State submenu ---------------------------------------------------------------
	my $powerstate_menu = Gtk2::Menu->new();

	my $menu_wake_up = Gtk2::ImageMenuItem->new_with_label( "Wake-Up / Power Button" );
	$menu_wake_up->signal_connect( activate => sub{remote_command('power');} );
	$powerstate_menu->append( $menu_wake_up );

	my $menu_standby = Gtk2::ImageMenuItem->new_with_label( "Standby" );
	$menu_standby->signal_connect( activate => sub{powerstate_command('standby');} );
	$powerstate_menu->append( $menu_standby );

	my $menu_deep_standby = Gtk2::ImageMenuItem->new_with_label( "Deep Standby" );
	$menu_deep_standby->signal_connect( activate => sub{powerstate_command('deep-standby');} );
	$powerstate_menu->append( $menu_deep_standby );

	my $menu_reboot = Gtk2::ImageMenuItem->new_with_label( "Reboot" );
	$menu_reboot->signal_connect( activate => sub{powerstate_command('reboot');} );
	$powerstate_menu->append( $menu_reboot );

	my $menu_restart_enigma = Gtk2::ImageMenuItem->new_with_label( "Restart Enigma 2" );
	$menu_restart_enigma->signal_connect( activate => sub{powerstate_command('restart-enigma');} );
	$powerstate_menu->append( $menu_restart_enigma );
	# -----------------------------------------------------------------------------------

	my $menu_powerstate = Gtk2::ImageMenuItem->new_with_label( "Power State" );
	$menu_powerstate->set_image( Gtk2::Image->new_from_icon_name( 'system-shutdown', 'menu' ) );
	$menu_powerstate->set_submenu( $powerstate_menu );
	$menu->append( $menu_powerstate );
	# -----------------------------------------------------------------------------------

	# Remote Control submenu ------------------------------------------------------------
	my $remotecontrol_menu = Gtk2::Menu->new();

	for my $rmcmd (sort keys %remote_codes) {
		(my $rmcmd_label = $rmcmd) =~ s/(?<!\w)(\w)/\u$1/g;
		my $menu_rmcommand = Gtk2::ImageMenuItem->new_with_label( $rmcmd_label );
		$menu_rmcommand->signal_connect( activate => sub{remote_command( $rmcmd );} );
		$remotecontrol_menu->append( $menu_rmcommand );
	}

	my $menu_remotecontrol = Gtk2::ImageMenuItem->new_with_label( "Remote Control" );
	$menu_remotecontrol->set_image( Gtk2::Image->new_from_stock( 'gtk-execute', 'menu' ) );
	$menu_remotecontrol->set_submenu( $remotecontrol_menu );
	$menu->append( $menu_remotecontrol );

	# -----------------------------------------------------------------------------------

	$menu->append(Gtk2::SeparatorMenuItem->new);

	my $menu_quit = Gtk2::ImageMenuItem->new_with_label( "Quit" );
	$menu_quit->signal_connect( activate => \&quit_app );
	$menu_quit->set_image( Gtk2::Image->new_from_stock( 'gtk-quit', 'menu' ) );
	$menu->append( $menu_quit );

	$menu->show_all;

	$menu->popup( undef, undef, undef, undef, 0, 0 );
	return 1;
}

sub update_channel_window_treeview {
	my $i=0;
	for my $s ( @services_index ) {
		my ($eventtitle, $runtime_info);
		if ( $s->{'e2eventtitle'} eq 'None' ) {
			$eventtitle = "";
			$runtime_info = "";
		} else {
			my $runtime = $s->{'e2eventcurrenttime'} - $s->{'e2eventstart'};
			my $percent = sprintf("%.0f", 100 / $s->{'e2eventduration'} * $runtime);
			my $s_remain = $s->{'e2eventduration'} - $runtime;
			my $m_remain = sprintf("%.0f", $s_remain / 60);
			$eventtitle = $s->{'e2eventtitle'};
			$runtime_info = "+" . $m_remain . " min (%" . $percent . ")";
		}
		# Don't recreate the array but overwrite each entry to prevent reseting the listview selection and scroll position
		${$channel_window_treeview->{data}}[$i] = [ ($i+1), $s->{'e2eventservicename'}, $eventtitle, $runtime_info ];
		$i++;
	}
	# Get rid of tailing entries if new channel list is shorter
	if ( scalar(@{$channel_window_treeview->{data}}) > $i ) {
		debug("Channel list length just got shorter, truncating treeview list from " . splice(@{$channel_window_treeview->{data}}) . " to " . $i, 1);
		splice( @{$channel_window_treeview->{data}}, $i);
	}
}

sub set_channel_number {
	my ($channel) = @_;

	return if $channel !~ /^\d{1,3}$/;

	my $reference = $services_index[$channel - 1]->{'e2eventservicereference'};

	if (! defined($reference) ) {
		debug("WARNING: channel number $channel has no reference.", 1);
		return;
	}

	debug("Switching to channel: $services_index[$channel - 1]->{'e2eventservicename'} (${channel}).", 1);
	get_xml_from_e2cmd( 'zap?sRef=' . $reference );
}

sub remote_channel_number {
	my ($channel) = @_;

	debug("Zap to channel: ${channel}", 1);

	return if $channel !~ /^\d{1,3}$/;

	foreach my $digit (split(//, $channel)) {	remote_command($digit) };
	remote_command('ok');
}

sub zap_channel_number {
	my ($channel) = @_;

	return if ($current_channel == $channel || $pwrstate ne 'running');

	if ( $zapmethod eq 'remote' ) {
		remote_channel_number($channel);
	} else {
		set_channel_number($channel);
	}
	$current_channel = $channel;
}

sub next_channel {
	my $channel = $current_channel + 1;
	if ( $channel > scalar @services_index ) {
		$channel = 1;
	}
	zap_channel_number($channel);
}

sub previous_channel {
	my $channel = $current_channel - 1;
	if ( $channel < 1 ) {
		$channel = scalar @services_index;
	}
	zap_channel_number($channel);
}

sub get_location {
	my $file = $_[0];
	my $a_searchpath = $_[1];

	for my $path (@{$a_searchpath}) {
		if ( -r "${path}/${file}" ) {
			debug("Selected file \"$file\" from location \"${path}/${file}\"", 1);
			return "${path}/${file}";
		}
	}
	return undef;
}

sub daemonize {
	open(SELFLOCK, "<$0") || abort("Couldn't open $0: $!");
	flock(SELFLOCK, LOCK_EX | LOCK_NB) or abort("Another daemon is already running");
	if ($debug > 0) {
		open(STDOUT, ">${startupdir}/debug-out");
	} else {
		open(STDOUT, ">/dev/null");
	}
	open(STDIN, "/dev/null");
	open(STDERR, ">&STDOUT");
	chdir('/');
	# Double-fork to avoid leaving a zombie process behind:
	exit if (fork());
	exit if (fork());
	sleep 1 until getppid() == 1;
}

sub abort {
	my ($msg) = @_;

	if ($msg) {
		$msg = "ABORT: ${msg}\n";
	} else {
		$msg = "ABORT: terminated.\n";
	}
	print STDERR "${msg}";

	exit 1;
}

sub debug {
	my ($msg, $level) = @_;
	$level = 1 if !$level;
	print "DEBUG: $msg\n" if $level <= $debug;
}

sub quit_app {
	Gtk2->main_quit;
}

sub display_help {
	(my $basename = $0) =~ s/.*\///;
	print "Usage: $basename [-f] [-d level] [-t timeout] [-h] remote address\n";
	print "\n";
	print "        -f             - Run in foregound (default:off)\n";
	print "        -d level       - Debug level 0-9 (default:0)\n";
	print "        -t timeout     - Set HTTP timeout in seconds 5-999 (default:10)\n";
	print "        -v             - Display version\n";
	print "        -h             - Display this help information\n";
	print "        -n             - Disable proxy (default:use http_proxy env)\n";
	print "        remote address - Hostname of the device running the Enigma 2 interface (required)\n";
	print "\n";
	exit 0;
}

sub open_audio_window {
	debug("Audio window requested.", 1);

	return if ($pwrstate ne 'running');

	my $xmltracklist = get_xml_from_e2cmd( 'getaudiotracks' );
	return if (!$xmltracklist->{'e2audiotrack'});

	my @tracklist;
	if ( ref($xmltracklist->{'e2audiotrack'} ) eq 'HASH' ) {
		$tracklist[0] = $xmltracklist->{'e2audiotrack'};
	} else {
		@tracklist = @{$xmltracklist->{'e2audiotrack'}};
	}

	my $audio_window = Gtk2::Window->new( 'toplevel' );
	$audio_window->set_position( 'mouse' );
	$audio_window->set_title ( 'Audio Tracks' );
	$audio_window->set_default_icon_from_file($iconfile);
	$audio_window->set_decorated(0);

	my $audio_window_root_vbox = Gtk2::VBox->new( 0, 6);
	$audio_window->add($audio_window_root_vbox);

	my $audio_window_frame = Gtk2::Frame->new();
	$audio_window_root_vbox->pack_start($audio_window_frame, TRUE, TRUE, 0);

	my $audio_window_vbox = Gtk2::VBox->new( 0, 6);
	$audio_window_frame->add($audio_window_vbox);

	my $audio_window_radio_frame = Gtk2::Frame->new('Audio Tracks');
	$audio_window_vbox->pack_start($audio_window_radio_frame, TRUE, TRUE, 0);
	$audio_window_vbox->set_border_width(4);

	# Radio Buttons -----------------------------------------------------------------
	my $audio_window_radio_vbox = Gtk2::VBox->new(0, 2);
	$audio_window_radio_vbox->set_border_width(5);
	$audio_window_radio_frame->add($audio_window_radio_vbox);

	my $group;
	my $i=0;

	for my $track (@tracklist) {
		my $button;

		$button = Gtk2::RadioButton->new($group, "$track->{'e2audiotrackid'} - $track->{'e2audiotrackdescription'}" );
		$group = $button->get_group if ($i == 0);

		$audio_window_radio_vbox->pack_start($button, TRUE, TRUE, 0);

		if ( $track->{'e2audiotrackactive'} eq 'True' ) {
			$button->set_active(TRUE);
		}
		$button->signal_connect( clicked => \&radio_click,$i );
		$i++;
	}

	sub radio_click {
		my ($button, $number) = @_;
		return if (!$button->get_active()); # Ignore click events triggered by button deactivations
		my $label = $button->get_label;
		debug("Audio track \"$label\" ($number) selected.", 1);
		get_xml_from_e2cmd( 'selectaudiotrack?id=' . $number);
	}

	# Close Button ------------------------------------------------------------------
	my $audio_window_close_hbox= Gtk2::HBox->new( 0, 6 );

	my $close_button = Gtk2::Button->new_from_stock('gtk-close');
	$close_button->set_border_width(2);
	$audio_window_close_hbox->pack_end( $close_button, 0, 0, 4 );
	$close_button->signal_connect( clicked => sub{ $audio_window->destroy; } );

	$audio_window_vbox->pack_start($audio_window_close_hbox,0,0,2);
	# -------------------------------------------------------------------------------

	$audio_window->signal_connect(delete_event => sub {	$audio_window->destroy; return 1; });

	$audio_window->signal_connect (key_press_event => sub {
			my ($widget, $event) = @_;
			return  unless $event->keyval == $Gtk2::Gdk::Keysyms{KEY_Escape};
			$audio_window->destroy;
			return 1;
		});

	$audio_window->show_all;
}
